// 002 u_c_cds_analyze_xml
// 16 oct 2004

// -- (C) John COLIBRI 2004
// -- documentation: http://www.jcolibri.com

(*$r+*)

unit u_c_net_cds_analyze_xml;
  interface
    uses Classes
        , u_c_net_text_buffer
        , u_c_net_cds_table
        ;

    type c_cds_analyze_xml= class(c_text_buffer)
                              m_c_result_list: tStringList;
                              m_trace_xml: Boolean;
                              m_file_name: String;

                              constructor create_cds_analyze_xml(p_name, p_file_name: String);
                              procedure analyze_xml(p_c_table: c_table);
                              destructor Destroy; Override;
                            end; // c_cds_analyze_xml

  implementation
    uses SysUtils, u_net_types_constants
        , u_net_strings
        , u_c_net_display
        ;

    constructor c_cds_analyze_xml.create_cds_analyze_xml(p_name, p_file_name: String);
      begin
        Inherited create_text_buffer(p_name, 0);
        m_file_name:= p_file_name;
      end; // create_cds_analyze_xml

    procedure c_cds_analyze_xml.analyze_xml(p_c_table: c_table);
        // -- analyze_xml:
        // --   <?> as is
        // --   <xxx> </xxx> : like begin end
        // --   if finds </yyy> bebore </xxx> goes one level up
      var l_result_line: String;
          l_indentation: Integer;

          l_row_number: Integer;
          // -- analyze the xml history
          l_c_merge_log_list: tStringList;
          l_c_merge_log_line_number_list: tStringList;

      const k_blanks= [' ', chr(9)];
            k_identifier_characters= ['a'..'z'];
            k_returns= [k_return, k_line_feed];

            k_tag_end= [' ', '>', chr(13)];
            k_other_end= k_returns+ ['<'];

      procedure trace_xml(p_text: String);
        begin
          if m_trace_xml
            then display(p_text);
        end; // trace_xml

      procedure add_text(p_text: String);
        begin
          l_result_line:= l_result_line+ p_text;
        end; // add_text

      procedure add_line;
        var l_blank_line: Boolean;
            l_index: Integer;
        begin
          // -- only add_text if not filled with spaces
          l_blank_line:= True;
          for l_index:= 1 to Length(l_result_line) do
            if l_result_line[l_index]<> ' '
              then begin
                  l_blank_line:= False;
                  Break;
                end;

          if not l_blank_line
            then m_c_result_list.Add(l_result_line);

          l_result_line:= '';
        end; // add_line

      function f_get_symbol_recursive: String;

        procedure get_number;
          var l_start_index: Integer;
              l_number: String;
          begin
            l_start_index:= m_buffer_index;

            if m_oa_text_buffer[m_buffer_index]= '-'
              then inc(m_buffer_index);
            while (m_buffer_index< m_buffer_size)
                and (m_oa_text_buffer[m_buffer_index] in k_digits) do
              inc(m_buffer_index);

            l_number:= f_extract_string_start_end(l_start_index, m_buffer_index- 1);
            add_text(l_number);
          end; // get_number

        procedure handle_opening_bracket(var pv_exit: Boolean);

          procedure skip_apostroph;
            var l_start_index, l_length: Integer;
                l_litteral: String;
            begin
              l_litteral:= '"';
              inc(m_buffer_index);

              l_start_index:= m_buffer_index;
              repeat
                inc(m_buffer_index);
              until (m_buffer_index>= m_buffer_size)
                  or (m_oa_text_buffer[m_buffer_index]= '"');

              l_length:= m_buffer_index- l_start_index;
              if l_length> 0
                then begin
                    l_litteral:= f_extract_string_start_end(l_start_index, m_buffer_index- 1);

                    // -- skip closing "
                    inc(m_buffer_index);
                  end
                else begin
                    display('*** did not find closing "');
                    inc(m_buffer_index);
                  end;
            end; // skip_apostroph

          procedure analyze_change_log_params(p_params: String);
            var l_line_size, l_line_index: Integer;
                l_key: String;

            procedure skip_blanks;
              begin
                while (l_line_index<= l_line_size) and (p_params[l_line_index] in [' ']) do
                  Inc(l_line_index);
              end; // skip_blanks

            procedure find_key;
              var l_start_index: Integer;
              begin
                l_start_index:= l_line_index;
                while (l_line_index<= l_line_size) and (p_params[l_line_index] in ['A'..'Z', '_']) do
                  Inc(l_line_index);

                l_key:= f_extract_string(p_params, l_start_index, l_line_index- 1);
              end; // find_key

            function f_integer: Integer;
              var l_start_index, l_length: Integer;
                  l_number: String;
              begin
                skip_blanks;

                l_start_index:= l_line_index;
                while (l_line_index<= l_line_size) and (p_params[l_line_index] in ['0'..'9']) do
                  Inc(l_line_index);

                l_length:= l_line_index- l_start_index;
                if l_length> 0
                  then begin
                      l_number:= f_extract_string(p_params, l_start_index, l_line_index- 1);
                      Result:= StrToInt(l_number);
                    end
                  else Result:= -1;
              end; // f_integer

            var l_delta_row, l_previous_row, l_code: Integer;

            begin // analyze_change_log_params
              trace_xml('anal_params');
              // -- "<PARAMS CHANGE_LOG="1 0 4 2 0 4 3 0 4 4 0 4 5 0 4"/>"
              trace_xml(p_params);

              l_line_size:= Length(p_params);
              l_line_index:= 1;

              skip_blanks;
              find_key;

              if l_key= 'CHANGE_LOG'
                then begin
                    Inc(l_line_index, 2);
                    while (l_line_index<= l_line_size) and (p_params[l_line_index]<> '"') do
                    begin
                      // -- get each triplet
                      l_delta_row:= f_integer;
                      l_previous_row:= f_integer;
                      l_code:= f_integer;
                      // -- display('  '+ IntToStr(l_delta_row)+ ' '+ IntToStr(l_previous_row)+ ' '+ IntToStr(l_code));
                      l_c_merge_log_list.Add(IntToStr(l_delta_row)+ ' '+ IntToStr(l_previous_row)+ ' '+ IntToStr(l_code));
                      l_c_merge_log_line_number_list.Add(IntToStr(l_delta_row));
                     end;
                  end;
            end; // analyze_change_log_params

          procedure analyze_tag(p_tag, p_parameters: String);
            var l_row_string, l_log_list_string: String;
                l_row_index: Integer;
            begin
              add_line;

              if p_tag= 'ROW'
                then begin
                    l_row_string:= f_spaces(l_indentation);
                    l_row_string:= l_row_string+ IntToStr(l_row_number+ 1);

                    // -- find if a CHANGE_LOG entry has this number
                    l_row_index:= l_c_merge_log_line_number_list.IndexOf(IntToStr(l_row_number+ 1));

                    If l_row_index<> -1
                      then l_log_list_string:= l_c_merge_log_list[l_row_index]
                      else l_log_list_string:= '';
                    l_row_string:= l_row_string+ ' ('+ l_log_list_string+ ') :';

                    l_row_string:= l_row_string+ ' <'+ p_tag+ p_parameters+ '>';
                    add_text(l_row_string);

                    // -- also add_text to the c_table structure
                    if p_c_table<> Nil
                      then p_c_table.f_c_add_line(l_row_number, l_log_list_string, p_parameters);

                    Inc(l_row_number);
                  end
                else add_text(f_spaces(l_indentation)+ '<'+ p_tag+ p_parameters+ '>');
              add_line;

              if (p_tag<> 'ROW') and (p_tag<> 'FIELD')
                then Inc(l_indentation, 2);

              add_text(f_spaces(l_indentation));

              // display('> analyze_tag '+ p_tag);
              // -- recurse
              Result:= f_get_symbol_recursive;

              // display('waits for /'+ p_tag);
              if Result<> '/'+ p_tag
                then begin
                    // -- backup
                    // display('*** /'+ p_tag+ ' not found but: '+ Result+ '<');
                    pv_exit:= True;
                    if (p_tag<> 'ROW') and (p_tag<> 'FIELD')
                      then Dec(l_indentation, 2);
                  end
                else begin
                    add_line;
                    Dec(l_indentation, 2);
                    // -- ?? assume that the </xxx> has never any params
                    add_text(f_spaces(l_indentation)+ '<'+ Result+ '>');
                    add_line;

                    add_text(f_spaces(l_indentation));

                    // -- look for the next symbol or tag
                    Result:= f_get_symbol_recursive;
                  end;

              // display('< analyze_tag '+ p_tag);
            end; // analyze_tag

          var l_start_index, l_length: Integer;
              l_slash: String;
              l_tag, l_parameters: String;

          begin // handle_opening_bracket
            trace_xml('> bracket');

            pv_exit:= False;

            // skip <
            inc(m_buffer_index);

            if m_oa_text_buffer[m_buffer_index]= '/'
              then begin
                  l_slash:= '/';
                  inc(m_buffer_index);
                end
              else l_slash:= '';

            l_start_index:= m_buffer_index;
            repeat
              inc(m_buffer_index);
            until (m_buffer_index>= m_buffer_size)
                or (m_oa_text_buffer[m_buffer_index] in k_tag_end);

            l_length:= m_buffer_index- l_start_index;
            if l_length> 0
              then begin
                  l_tag:= f_extract_string_start_end(l_start_index, m_buffer_index- 1);
                  l_tag:= l_slash+ UpperCase(l_tag);

                  // display('tag ='+ l_tag+ '=');
                  Result:= l_tag;

                  // -- get command parameters
                  if m_oa_text_buffer[m_buffer_index]<> '>'
                    then begin
                        l_parameters:= '';

                        l_start_index:= m_buffer_index;
                        repeat
                          if m_oa_text_buffer[m_buffer_index]= '"'
                            then skip_apostroph
                            else inc(m_buffer_index);
                        until (m_buffer_index>= m_buffer_size)
                            or (m_oa_text_buffer[m_buffer_index] in ['>']);

                        l_parameters:= f_extract_string_start_end(l_start_index, m_buffer_index- 1);

                        if l_tag= 'PARAMS'
                          then analyze_change_log_params(l_parameters);
                      end; // parameters

                  // -- skip >
                  if m_oa_text_buffer[m_buffer_index]= '>'
                    then begin
                        inc(m_buffer_index);

                        if l_slash= '/'
                          then Result:= l_tag
                          else analyze_tag(l_tag, l_parameters);
                      end
                    else begin
                        display('*** could not find >');
                      end;
                end
              else begin
                  l_tag:= '';
                  display('*** tag is void');
                end;

            trace_xml('< bracket');
          end; // handle_opening_bracket

        procedure handle_return;
          begin
            repeat
              if (m_buffer_index< m_buffer_size)
                  and (m_oa_text_buffer[m_buffer_index]= k_return)
                then begin
                    add_text(' ');

                    inc(m_buffer_index);

                    if (m_buffer_index< m_buffer_size)
                        and (m_oa_text_buffer[m_buffer_index]= k_line_feed)
                      then inc(m_buffer_index);
                  end;
            until (m_buffer_index>= m_buffer_size) or (m_oa_text_buffer[m_buffer_index]<> k_return);
          end; // handle_return

        procedure handle_line_feed;
          begin
            display_bug_stop('LF');
          end; // handle_line_feed

        const k_separators= [chr(13), '{', '}', '\'];

        procedure handle_other;
          var l_start_index: Integer;
              l_other: String;
          begin
            l_start_index:= m_buffer_index;
            repeat
              inc(m_buffer_index);
            until (m_buffer_index>= m_buffer_size)
                or (m_oa_text_buffer[m_buffer_index] in k_other_end);

            l_other:= f_extract_string_start_end(l_start_index, m_buffer_index- 1);
            add_text(l_other);
          end; // handle_other

        var l_exit: Boolean;

        begin // f_get_symbol_recursive
          Result:= '';
          if m_buffer_index>= m_buffer_size
            then exit;

          repeat
            case m_oa_text_buffer[m_buffer_index] of
              '<' : begin
                      handle_opening_bracket(l_exit);
                      if l_exit
                        then exit;
                    end;
              chr(13) : handle_return;
              chr(10): handle_line_feed;

              else handle_other;
            end; // case
          until (m_buffer_index>= m_buffer_size) or (Result<> '');
        end; // f_get_symbol_recursive

      var l_symbol: String;

      begin  // analyze_xml
        trace_xml('do analyze_xml');

        if f_load_from_file(m_file_name)
          then begin
              m_c_result_list:= tStringList.Create;
              l_result_line:= '';

              // display('size '+ IntToStr(m_buffer_size));
              l_indentation:= 1;
              l_row_number:= 0;
              l_c_merge_log_list:= tStringList.Create;
              l_c_merge_log_line_number_list:= tStringList.Create;

              while not f_end_of_text do
                l_symbol:= f_get_symbol_recursive;

              add_line;
            end // load ok
          else display('*** could not open '+ m_file_name);
      end; // analyze_xml

    destructor c_cds_analyze_xml.Destroy;
      begin
        m_c_result_list.Free;

        Inherited Destroy;
      end; // Destroy

    begin // u_xml_1
    end. // u_xml_1


