// 001 u_c_line
// 03 mar 00

// -- also used by ed

// -- a line

// -- has line_number and buffer_index_of_start for ed or c_text_file,
// --   but not used in this unit

(*$r+*)

unit u_c_net_line;
  interface
    uses u_c_net_basic_object, u_net_types_constants;

    type c_line= class(c_basic_object)
                   m_the_line: String;
                   m_line_index, m_line_size: Word;

                   m_line_line_number: Word;
                   m_buffer_index_of_start: Integer;

                   m_extract_error: Integer;

                   Constructor create_line(p_name: String); Virtual;
                   function f_extract_string_start_length(p_start, p_length: Integer): String;
                   function f_extract_string_start_end(p_start, p_end: Integer): String;
                   procedure set_line(p_line: String);

                   procedure skip_blanks;
                   function f_read_integer: Integer;
                   function f_read_double: Double;
                   function f_read_non_blank_symbol: String;

                   destructor Destroy; Override;
                 end;

  implementation
    uses SysUtils
         , u_net_strings
        , u_c_net_display
    ;

    constructor c_line.create_line(p_name: String);
      begin
        Inherited create_basic_object(p_name);
      end; // Create

    function c_line.f_extract_string_start_length(p_start, p_length: Integer): String;
      begin
        m_extract_error:= 0;
        if p_length= 0
          then Result:= ''
          else begin
              if (p_length< 0) or (p_start< 1) or (p_start+ p_length- 1> m_line_size)
                then begin
                    if p_length< 0
                      then display('*** Length '+ IntToStr(p_length));
                    if p_start< 1
                      then display_bug('start '+ IntToStr(p_start));
                    if p_start+ p_length- 1> m_line_size
                      then display_bug('after_end '+ IntToStr(p_start)
                          + ' '+ IntToStr(p_length)+ ' '+ IntToStr(m_line_size));
                    m_extract_error:= 1;
                    display(m_the_line);
                    display(f_spaces(p_start- 1)+ '^');
                    display(f_spaces(p_start+ p_length- 1)+ '|');
                    display(f_spaces(m_line_size- 1)+ '^');
                    // stop(
                    display('*** m_c_line.f_extract_string_start_length '
                        // + m_c_file_name.m_complete_name+ ' '
                        + IntToStr(p_start)+ ' '+ IntToStr(p_length)+ ' '+ IntToStr(m_line_size)
                        + ' '+ IntToStr(p_start+ p_length));

                    // Result:= '';
                    Result:= m_the_line+ chr(13)+ chr(10)
                        + '*** m_c_line.f_extract_string_start_length '
                        // + m_c_file_name.m_complete_name+ ' '
                        + IntToStr(p_start)+ ' '+ IntToStr(p_length)+ ' '+ IntToStr(m_line_size)
                        + ' '+ IntToStr(p_start+ p_length);
                  end
                else begin
                    Result:= Copy(m_the_line, p_start, p_length);
                    // SetLength(Result, p_length);
                    // Move(m_the_line[p_start], Result[1], p_length);
                  end;
            end;
      end; // f_extract_string_start_length

    function c_line.f_extract_string_start_end(p_start, p_end: Integer): String;
        // -- [start..end]
      begin
        Result:= f_extract_string_start_length(p_start, p_end+ 1- p_start)
      end; // f_extract_string_start_end

    procedure c_line.set_line(p_line: String);
      begin
        m_the_line:= p_line;
        m_line_size:= Length(m_the_line);
        m_line_index:= 1;
      end; // set_line

    procedure c_line.skip_blanks;
      const k_blanks= [' ', k_tabulation];
      begin
        while (m_line_index<= m_line_size) and (m_the_line[m_line_index] in k_blanks) do
          Inc(m_line_index);
      end; // skip_blanks

    function c_line.f_read_integer: Integer;
      const k_number= ['0'..'9', '+', '-'];
      var l_start_index: Integer;
         l_string: String;
      begin
        l_start_index:= m_line_index;
        while (m_line_index<= m_line_size) and (m_the_line[m_line_index] in k_number) do
          Inc(m_line_index);
        l_string:= f_extract_string_start_end(l_start_index, m_line_index- 1);
        Result:= StrToInt(l_string);
      end; // f_read_integer:

    function c_line.f_read_double: Double;
      const k_number= ['0'..'9', '+', '-', '.', 'e'];
      var l_start_index: Integer;
          l_string: String;
      begin
        l_start_index:= m_line_index;
        while (m_line_index<= m_line_size) and (m_the_line[m_line_index] in k_number) do
          Inc(m_line_index);
        l_string:= f_extract_string_start_end(l_start_index, m_line_index- 1);
        // DecimalSeparator:= '.';
        Result:= StrToFloat(l_string);
      end; // f_read_double

    function c_line.f_read_non_blank_symbol: String;
      var l_start_index: Integer;
      begin
        skip_blanks;
        l_start_index:= m_line_index;
        while (m_line_index<= m_line_size) and not (m_the_line[m_line_index] in k_blanks) do
          Inc(m_line_index);

        Result:= f_extract_string_start_end(l_start_index, m_line_index- 1);
      end; // f_read_non_blank_symbol

    destructor c_line.Destroy;
      begin
        Inherited;
      end; // Destroy

    begin // u_c_line
    end. // u_c_line
